@@ -7,12 +7,14 @@ from .utils import is_integer_value, _compute_bitshift, WIDE_MASK
 from .utils import check_sentinel
 from .healSparseCoverage import HealSparseCoverage
 from .fits_shim import _make_header
+from .packedBoolArray import _PackedBoolArray
 
 use_pyarrow = False
 try:
     import pyarrow as pa
     from pyarrow import parquet
     from pyarrow import dataset
+    from .parquet_shim import to_numpy_dtype
     use_pyarrow = True
 except ImportError:
     pass
@@ -71,7 +73,6 @@ def _read_map_parquet(healsparse_class, filepath, pixels=None, header=False,
         # Note that this could be reconstructed from the information in the file
         # inefficiently.  This feature could be added in the future.
         raise RuntimeError("Filepath %s is missing coverage map %s" % (filepath, cov_fname))
-
     nside_sparse = int(md['healsparse::nside_sparse'])
     nside_coverage = int(md['healsparse::nside_coverage'])
     nside_io = int(md['healsparse::nside_io'])
@@ -111,18 +112,25 @@ def _read_map_parquet(healsparse_class, filepath, pixels=None, header=False,
         is_wide_mask = False
         wmult = 1
 
+    if md['healsparse::bitpacked'] == 'True':
+        is_bit_packed = True
+        wdiv = 8
+    else:
+        is_bit_packed = False
+        wdiv = 1
+
     if md['healsparse::primary'] != '':
         # This is a multi-column table.
         is_rec_array = True
         primary = md['healsparse::primary']
         columns = [name for name in schema.names if name not in ['iopix', 'cov_pix']]
-        dtype = [(name, schema.field(name).type.to_pandas_dtype()) for
+        dtype = [(name, to_numpy_dtype(schema.field(name).type)) for
                  name in columns]
-        primary_dtype = schema.field(primary).type.to_pandas_dtype()
+        primary_dtype = to_numpy_dtype(schema.field(primary).type)
     else:
         is_rec_array = False
         primary = None
-        dtype = schema.field('sparse').type.to_pandas_dtype()
+        dtype = to_numpy_dtype(schema.field('sparse').type)
         primary_dtype = dtype
         columns = ['sparse']
 
@@ -150,8 +158,8 @@ def _read_map_parquet(healsparse_class, filepath, pixels=None, header=False,
                 continue
             sparse_map[d[0]][: cov_map.nfine_per_cov] = check_sentinel(d[1], None)
     else:
-        sparse_map = np.zeros((_pixels.size + 1)*cov_map.nfine_per_cov*wmult, dtype=dtype)
-        sparse_map[: cov_map.nfine_per_cov*wmult] = sentinel
+        sparse_map = np.zeros((_pixels.size + 1)*cov_map.nfine_per_cov*wmult//wdiv, dtype=dtype)
+        sparse_map[: cov_map.nfine_per_cov*wmult//wdiv] = sentinel
 
     if _pixels_io is None:
         # Read the full table
@@ -175,11 +183,13 @@ def _read_map_parquet(healsparse_class, filepath, pixels=None, header=False,
         for name in columns:
             sparse_map[name][cov_map.nfine_per_cov:] = tab[name].to_numpy()
     else:
-        sparse_map[cov_map.nfine_per_cov*wmult:] = tab['sparse'].to_numpy()
+        sparse_map[cov_map.nfine_per_cov*wmult//wdiv:] = tab['sparse'].to_numpy()
 
         if is_wide_mask:
             sparse_map = sparse_map.reshape((sparse_map.size // wmult,
                                              wmult)).astype(WIDE_MASK)
+        elif is_bit_packed:
+            sparse_map = _PackedBoolArray(data_buffer=sparse_map)
 
     healsparse_map = healsparse_class(cov_map=cov_map, sparse_map=sparse_map,
                                       nside_sparse=nside_sparse, primary=primary,
@@ -245,6 +255,11 @@ def _write_map_parquet(hsp_map, filepath, clobber=False, nside_io=4):
     else:
         wmult = 1
 
+    if hsp_map.is_bit_packed_map:
+        wdiv = 8
+    else:
+        wdiv = 1
+
     if np.isclose(hsp_map._sentinel, hpg.UNSEEN):
         sentinel_string = 'UNSEEN'
     else:
@@ -258,17 +273,23 @@ def _write_map_parquet(hsp_map, filepath, clobber=False, nside_io=4):
                 'healsparse::primary': '' if hsp_map.primary is None else hsp_map.primary,
                 'healsparse::sentinel': sentinel_string,
                 'healsparse::widemask': str(hsp_map.is_wide_mask_map),
-                'healsparse::wwidth': str(hsp_map._wide_mask_width)}
+                'healsparse::wwidth': str(hsp_map._wide_mask_width),
+                'healsparse::bitpacked': str(hsp_map.is_bit_packed_map)}
 
     # Add additional metadata
     if hsp_map.metadata is not None:
         # Use the fits header serialization for compatibility
-        hdr_string = _make_header(hsp_map.metadata).tostring()
+        hdr_string = str(_make_header(hsp_map.metadata, force_astropy=True))
         metadata['healsparse::header'] = hdr_string
 
+    if hsp_map.is_bit_packed_map:
+        sparse_map = hsp_map._sparse_map.data_array
+    else:
+        sparse_map = hsp_map._sparse_map.ravel()
+
     if not hsp_map.is_rec_array:
         schema = pa.schema([('cov_pix', pa.from_numpy_dtype(np.int32)),
-                            ('sparse', pa.from_numpy_dtype(hsp_map.dtype))],
+                            ('sparse', pa.from_numpy_dtype(sparse_map.dtype))],
                            metadata=metadata)
     else:
         type_list = [(name, pa.from_numpy_dtype(hsp_map.dtype[name].type)) for
@@ -277,45 +298,45 @@ def _write_map_parquet(hsp_map, filepath, clobber=False, nside_io=4):
         schema = pa.schema(type_list, metadata=metadata)
 
     cov_map = hsp_map._cov_map
-    sparse_map = hsp_map._sparse_map.ravel()
     cov_index_map_temp = cov_map[:] + np.arange(hpg.nside_to_npixel(hsp_map.nside_coverage),
                                                 dtype=np.int64)*cov_map.nfine_per_cov
 
-    pix_arr = np.zeros(cov_map.nfine_per_cov*wmult, dtype=np.int32)
+    cpix_arr = np.zeros(cov_map.nfine_per_cov*wmult//wdiv, dtype=np.int32)
 
-    last_pix_io = -1
+    last_cpix_io = -1
     writer = None
     row_groups = np.zeros_like(cov_pixels)
-    for ctr, (pix_io, pix) in enumerate(zip(cov_pixels_io, cov_pixels)):
+    for ctr, (cpix_io, cpix) in enumerate(zip(cov_pixels_io, cov_pixels)):
         # These are always going to be sorted
-        if pix_io > last_pix_io:
-            last_pix_io = pix_io
+        if cpix_io > last_cpix_io:
+            last_cpix_io = cpix_io
 
             if writer is not None:
                 writer.close()
                 writer = None
 
             # Create a new file
-            pixpath = os.path.join(filepath, f'iopix={pix_io:03d}')
-            os.makedirs(pixpath)
+            iopixpath = os.path.join(filepath, f'iopix={cpix_io:03d}')
+            os.makedirs(iopixpath)
 
-            pixfile = os.path.join(pixpath, f'{pix_io:03d}.parquet')
-            writer = parquet.ParquetWriter(pixfile, schema)
+            iopixfile = os.path.join(iopixpath, f'{cpix_io:03d}.parquet')
+            writer = parquet.ParquetWriter(iopixfile, schema)
             row_group_ctr = 0
 
-        sparsepix = sparse_map[cov_index_map_temp[pix]*wmult:
-                               (cov_index_map_temp[pix] + cov_map.nfine_per_cov)*wmult]
-        pix_arr[:] = pix
+        sparsepix = sparse_map[cov_index_map_temp[cpix]*wmult//wdiv:
+                               (cov_index_map_temp[cpix] + cov_map.nfine_per_cov)*wmult//wdiv]
+        cpix_arr[:] = cpix
         if not hsp_map.is_rec_array:
-            arrays = [pa.array(pix_arr),
+            arrays = [pa.array(cpix_arr),
                       pa.array(sparsepix)]
         else:
             arrays = [pa.array(sparsepix[name]) for
                       name in hsp_map.dtype.names]
-            arrays[0: 0] = [pa.array(pix_arr)]
+            arrays[0: 0] = [pa.array(cpix_arr)]
         tab = pa.Table.from_arrays(arrays, schema=schema)
 
-        writer.write_table(tab)
+        # Ensure we write this as one row group.
+        writer.write_table(tab, row_group_size=len(tab))
         row_groups[ctr] = row_group_ctr
         row_group_ctr += 1
 
